@@ -3,6 +3,14 @@
 #include <ctype.h>
 #include <sys/stat.h>
 
+#ifndef S_ISDIR
+#define S_ISDIR(mode)  (((mode) & S_IFMT) == S_IFDIR)
+#endif
+
+#ifndef S_ISREG
+#define S_ISREG(mode)  (((mode) & S_IFMT) == S_IFREG)
+#endif
+
 static char msg[4069];
 /*
 #ifdef _DEFINE_WIN32_FSEEKO
@@ -18,6 +26,30 @@ static char msg[4069];
 #endif
 */
 
+
+int saprintf(char **retp, const char *fmt, ...) {
+  va_list argp;
+  int len;
+  char *buf;
+
+  va_start(argp, fmt);
+  len = vsnprintf(NULL, 0, fmt, argp);
+  va_end(argp);
+  GMALLOC(buf, (len + 1));
+  if(buf == NULL)
+    {
+    *retp = NULL;
+    return -1;
+    }
+
+  va_start(argp, fmt);
+  vsnprintf(buf, len+1, fmt, argp);
+  va_end(argp);
+
+  *retp = buf;
+  return len;
+}
+
 //************************* Debug helpers **************************
 // Assert failed routine
 void GAssert(const char* expression, const char* filename, unsigned int lineno){
@@ -47,6 +79,7 @@ void GError(const char* format,...){
   #endif
     exit(1);
   }
+  
 // Warning routine (just print message without exiting)
 void GMessage(const char* format,...){
   va_list arguments;
@@ -347,13 +380,13 @@ char* GLineReader::getLine(FILE* stream, off_t& f_pos) {
 
 
 //strchr but with a set of chars instead of only one
-char* strchrs(char* s, const char* chrs) {
+char* strchrs(const char* s, const char* chrs) {
   if (s==NULL || chrs==NULL || *chrs=='\0' || *s=='\0')
          return NULL;
   unsigned int l=strlen(s);
   unsigned int r=strcspn(s, chrs);
   if (r==l) return NULL;
-  return (s+r);
+  return ((char*)s+r);
 }
 
 char* upCase(const char* str) {
@@ -393,9 +426,9 @@ char* strupper(char * str) {//changes string in place
 
 
 //test if a char is in a given string (set)
-bool chrInStr(char c, char* str) {
+bool chrInStr(char c, const char* str) {
  if (str==NULL || *str=='\0') return false;
- for (char* p=str; (*p)!='\0'; p++) {
+ for (const char* p=str; (*p)!='\0'; p++) {
    if ((*p)==c) return true;
    }
  return false;
@@ -403,13 +436,13 @@ bool chrInStr(char c, char* str) {
 
 
 
-char* rstrfind(char* str, const char* substr) {
+char* rstrfind(const char* str, const char* substr) {
 /* like rindex() for a string */
  int l,i;
  if (str==NULL || *str=='\0') return NULL;
  if (substr==NULL || *substr=='\0') return NULL;
  l=strlen(substr);
- char* p=str+strlen(str)-l;
+ char* p=(char*)str+strlen(str)-l;
    //rightmost position that could match
 
  while (p>=str) {
@@ -421,15 +454,15 @@ char* rstrfind(char* str, const char* substr) {
 }
 
 
-char* strifind(char* str,  const char* substr) {
+char* strifind(const char* str,  const char* substr) {
  // the case insensitive version of strstr -- finding a string within a strin
   int l,i;
   if (str==NULL || *str==0) return NULL;
   if (substr==NULL || *substr==0) return NULL;
   l=strlen(substr);
-  char* smax=str+strlen(str)-l;
+  char* smax=(char*)str+strlen(str)-l;
   //rightmost position that could match
-  char* p=str;
+  char* p=(char*)str;
   while (p<=smax) {
      for (i=0; i<l && tolower(*(p+i))==tolower(*(substr+i)); i++) ;
      if (i==l) return p; //found!
@@ -441,36 +474,46 @@ char* strifind(char* str,  const char* substr) {
 
 
 // tests if string s has the given prefix
-bool startsWith(char* s, const char* prefix) {
+bool startsWith(const char* s, const char* prefix) {
  if (prefix==NULL || s==NULL) return false;
  int i=0;
  while (prefix[i]!='\0' && prefix[i]==s[i]) i++;
  return (prefix[i]=='\0');
  }
 
+// tests if string s ends with given suffix
+bool endsWith(const char* s, const char* suffix) {
+ if (suffix==NULL || s==NULL) return false;
+ if (suffix[0]==0) return true; //special case: empty suffix
+ int j=strlen(suffix)-1;
+ int i=strlen(s)-1;
+ if (i<j) return false;
+ while (j>=0 && s[i]==suffix[j]) { i--; j--; }
+ return (j==-1);
+ }
+
 
 char* reverseChars(char* str, int slen) {
   if (slen==0) slen=strlen(str);
   int l=0;
   int r=slen-1;
-  register char c;
+  char c;
   while (l<r) {
      c=str[l];str[l]=str[r];
      str[r]=c;
-     //swap(str[l],str[r]);
      l++;r--;
      }
   return str;
 }
 
 
-char* rstrstr(char* rstart, char *lend, char* substr) {  /*like strstr, but starts searching
+char* rstrstr(const char* rstart, const char *lend, const char* substr) {  /*like strstr, but starts searching
  from right end, going up to lend and returns a pointer to the last (right)
  matching character in str */
  char *p;
  int l,i;
  l=strlen(substr);
- p=rstart-l+1;
+ p=(char*)rstart-l+1;
  while (p>=lend) {
     for (i=0;i<l;i++) if (*(p+i) != *(substr+i)) break;
     if (i==l) return p+l-1;
@@ -494,26 +537,38 @@ int strhash(const char* str){
   return h;
   }
 
-// removes the directory part from a full-path file name
+// removes the last part (file or directory name) of a full path
 // this is a destructive operation for the given string!!!
 // the trailing '/' is guaranteed to be there
 void delFileName(char* filepath) {
  char *p, *sep;
  if (filepath==NULL) return;
  for (p=filepath, sep=filepath;*p!='\0';p++)
-     if (*p==CHPATHSEP) sep=p+1;
+     if (*p=='/' || *p=='\\') sep=p+1;
  *sep='\0'; // truncate filepath
 }
 
-// returns a pointer to the file name part in a full-path filename
-char* getFileName(char* filepath) {
- char *p, *sep;
+// returns a pointer to the last file or directory name in a full path
+const char* getFileName(const char* filepath) {
+ const char *p, *sep;
  if (filepath==NULL) return NULL;
  for (p=filepath, sep=filepath;*p!='\0';p++)
-     if (*p==CHPATHSEP) sep=p+1;
+     if (*p=='/' || *p=='\\') sep=p+1;
  return sep;
 }
 
+// returns a pointer to the file "extension" part in a filename
+const char* getFileExt(const char* filepath) {
+ const char *p, *dp, *sep;
+ if (filepath==NULL) return NULL;
+ for (p=filepath, dp=filepath, sep=filepath;*p!='\0';p++) {
+     if (*p=='.') dp=p+1;
+       else if (*p=='/' || *p=='\\') 
+                  sep=p+1;
+     }
+ return (dp>sep) ? dp : NULL ;
+}
+
 int fileExists(const char* fname) {
   struct stat stFileInfo;
   int r=0;
@@ -678,3 +733,28 @@ void writeFasta(FILE *fw, const char* seqid, const char* descr,
   fflush(fw);
  }
 
+char* commaprint(uint64 n) {
+  static int comma = '\0';
+  static char retbuf[48];
+  char *p = &retbuf[sizeof(retbuf)-1];
+  int i = 0;
+  if(comma == '\0') {
+    /* struct lconv *lcp = localeconv();
+    if(lcp != NULL) {
+      if(lcp->thousands_sep != NULL &&
+        *lcp->thousands_sep != '\0')
+        comma = *lcp->thousands_sep;
+      else  */
+                          comma = ',';
+     // }
+    }
+  *p = '\0';
+  do {
+    if(i%3 == 0 && i != 0)
+      *--p = comma;
+    *--p = '0' + n % 10;
+    n /= 10;
+    i++;
+  } while(n != 0);
+  return p;
+}
