@@ -73,7 +73,7 @@ void GStr::replace_data(Data *data) {
     my_data = data;
 }
 
-void GStr::make_unique() {//make sure is not a reference to other string
+void GStr::make_unique() {//make sure it's not a reference to other string
     if (my_data->ref_count > 1) {
         Data *data = new_data(length());
         ::memcpy(data->chars, chars(), length());
@@ -331,12 +331,29 @@ bool GStr::contains(const char *s) const {
  }
 
 bool GStr::startsWith(const char *s) const {
- return (index(s, 0) == 0);
+ //return (index(s, 0) == 0);
+ return ::startsWith(this->chars(), s);
+ }
+
+bool GStr::startsWith(const GStr& s) const {
+ //return (index(s, 0) == 0);
+ return ::startsWith(this->chars(), s.chars());
+ }
+
+bool GStr::endsWith(const char *s) const {
+ //return (index(s, 0) == 0);
+ return ::endsWith(this->chars(), s);
+ }
+
+bool GStr::endsWith(const GStr& s) const {
+ //return (index(s, 0) == 0);
+ return ::endsWith(this->chars(), s.chars());
  }
 
 bool GStr::contains(char c) const {
  return (index(c, 0) >= 0);
  }
+
 GStr& GStr::format(const char *fmt,...) {
 // Format as in sprintf
   make_unique(); //edit operation ahead
@@ -655,6 +672,20 @@ GStr GStr::substr(int idx, int len) const {
     return newstring;
 }
 
+GStr& GStr::reverse() {
+  make_unique();
+  int l=0;
+  int r=my_data->length-1;
+  char c;
+  while (l<r) {
+     c=my_data->chars[l];
+     my_data->chars[l]=my_data->chars[r];
+     my_data->chars[r]=c;
+     l++;r--;
+     }
+  return *this;
+}
+
 
 //transform: any character from 'from' is replaced with a coresponding
 //char from 'to'
