@@ -10,8 +10,14 @@
 #include "filters.h"
 #include <algorithm>
 #include <numeric>
+#include <boost/graph/adjacency_list.hpp>
+#include <boost/graph/depth_first_search.hpp>
+#include <boost/graph/visitors.hpp>
+#include <boost/graph/graph_traits.hpp>
+#include <boost/graph/connected_components.hpp>
 
 using namespace std;
+using namespace boost;
 
 void filter_introns(int bundle_length,
 					int bundle_left,
@@ -38,11 +44,11 @@ void filter_introns(int bundle_length,
 	{
 		bundle_avg_doc = major_isoform_intron_doc(intron_doc);
 		bundle_avg_thresh = fraction * bundle_avg_doc;
-		asm_verbose("\tFiltering bundle introns, avg (intron) doc = %lf, thresh = %f\n", bundle_avg_doc, bundle_avg_thresh);
+		verbose_msg("\tFiltering bundle introns, avg (intron) doc = %lf, thresh = %f\n", bundle_avg_doc, bundle_avg_thresh);
 	}
 	else
 	{
-		asm_verbose("\tFiltering bundle introns, avg bundle doc = %lf, thresh = %f\n", bundle_avg_doc, bundle_avg_thresh);
+		verbose_msg("\tFiltering bundle introns, avg bundle doc = %lf, thresh = %f\n", bundle_avg_doc, bundle_avg_thresh);
 	}
 	
 	for(map<pair<int, int>, int>::const_iterator itr = intron_doc.begin();
@@ -65,7 +71,7 @@ void filter_introns(int bundle_length,
 					if (doc < bundle_avg_thresh)
 					{
 						toss[j] = true;
-						asm_verbose("\t Filtering intron %d - %d: %f thresh %f\n", itr->first.first, itr->first.second, doc, bundle_avg_thresh);
+						verbose_msg("\t Filtering intron %d - %d: %f thresh %f\n", itr->first.first, itr->first.second, doc, bundle_avg_thresh);
 						continue; 
 					}
 					
@@ -86,7 +92,7 @@ void filter_introns(int bundle_length,
 						double thresh = itr2->second * fraction;
 						if (doc < thresh)
 						{
-							asm_verbose("\t Filtering intron (due to overlap) %d - %d: %f thresh %f\n", itr->first.first, itr->first.second, doc, bundle_avg_thresh);
+							verbose_msg("\t Filtering intron (due to overlap) %d - %d: %f thresh %f\n", itr->first.first, itr->first.second, doc, bundle_avg_thresh);
 							toss[j] = true;
 						}
 					}
@@ -100,7 +106,7 @@ void filter_introns(int bundle_length,
 		if (!toss[j])
 		{
 			filtered_hits.push_back(hits[j]);
-//#if ASM_VERBOSE
+//#if verbose_msg
 //			if (hits[j].has_intron())
 //			{
 //				fprintf(stderr, "KEEPING intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
@@ -112,13 +118,13 @@ void filter_introns(int bundle_length,
 			if (hits[j].has_intron())
 			{
 				
-				asm_verbose("\tFiltering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
+				verbose_msg("\tFiltering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
 			}
 		}
 	}
 	
 	
-	asm_verbose("\tIntron filtering pass finished: excluded %d fragments\n", (int)hits.size() - (int)filtered_hits.size());
+	verbose_msg("\tIntron filtering pass finished: excluded %d fragments\n", (int)hits.size() - (int)filtered_hits.size());
 	hits = filtered_hits;
 }
 
@@ -189,13 +195,13 @@ void pre_mrna_filter(int bundle_length,
                                                    i_right - bundle_left);
         
         double cumul_cov = 0;
-        for (size_t i = 0; i < i_right - i_left; ++i)
+        for (int i = 0; i < i_right - i_left; ++i)
         {
             size_t pos = (i_left - bundle_left) + i;
             cumul_cov += depth_of_coverage[pos];
         }
         cumul_cov /= i_right - i_left;
-        asm_verbose("retained intron %d-%d background: %lf\n", i_left, i_right, intron_background);
+        verbose_msg("retained intron %d-%d background: %lf\n", i_left, i_right, intron_background);
         if (cumul_cov / bundle_avg_doc >= pre_mrna_fraction)
         {
             //fprintf(stderr, "\tskipping\n");
@@ -203,12 +209,14 @@ void pre_mrna_filter(int bundle_length,
             continue;
         }
         
+		//double thresh = (1.0/pre_mrna_fraction) * intron_background;
+		double thresh = pre_mrna_fraction * intron_background;
+
         for (size_t j = 0; j < hits.size(); ++j)
         {
-            //if (hits[j].has_intron())
-            //    continue;
-            double thresh = (1.0/pre_mrna_fraction) * intron_background;
-            
+            if (hits[j].is_ref())
+                continue;
+			
             int len = 0;
             double doc = 0.0;
             size_t curr_op = 0;
@@ -289,166 +297,22 @@ void pre_mrna_filter(int bundle_length,
 		if (!toss[j])
 		{
 			filtered_hits.push_back(hits[j]);
-			//#if ASM_VERBOSE
-			//			if (hits[j].has_intron())
-			//			{
-			//				
-			//				fprintf(stderr, "KEEPING intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
-			//			}
-			//#endif	
 		}
 		else
 		{
 			if (hits[j].has_intron())
 			{
 				
-                asm_verbose( "\t@@@ Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
+                verbose_msg( "\t@@@ Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
 			}
 		}
 	}
     
-	asm_verbose("\tPre-mRNA filter pass complete, excluded %lu fragments\n", hits.size() - filtered_hits.size());
+	verbose_msg("\tPre-mRNA filter pass complete, excluded %lu fragments\n", hits.size() - filtered_hits.size());
 	
 	hits = filtered_hits;
 }
 
-//void pre_mrna_filter(int bundle_length,
-//					 int bundle_left,
-//					 vector<Scaffold>& hits)
-//{
-//	vector<int> depth_of_coverage(bundle_length,0);
-//	vector<double> scaff_doc;
-//	map<pair<int,int>, int> intron_doc;
-//	vector<Scaffold> filtered_hits;
-//	vector<bool> toss(hits.size(), false);
-//	
-//	// Make sure the avg only uses stuff we're sure isn't pre-mrna fragments
-//	double bundle_avg_doc = compute_doc(bundle_left, 
-//										hits, 
-//										depth_of_coverage, 
-//										intron_doc,
-//										true);
-//	
-//	// recompute the real DoCs
-//	compute_doc(bundle_left, 
-//				hits, 
-//				depth_of_coverage, 
-//				intron_doc,
-//				false);
-//	
-//	record_doc_for_scaffolds(bundle_left, 
-//							 hits, 
-//							 depth_of_coverage, 
-//							 intron_doc,
-//							 scaff_doc);
-//	
-////	vector<int>::iterator new_end = remove(depth_of_coverage.begin(), depth_of_coverage.end(), 0);
-////	depth_of_coverage.erase(new_end, depth_of_coverage.end());
-////	sort(depth_of_coverage.begin(), depth_of_coverage.end());
-////	
-////	size_t median = floor(depth_of_coverage.size() / 2);
-//	
-//	
-//	Scaffold smashed_gene;
-//	vector<Scaffold> spliced_hits;
-//    for (size_t i = 0; i < hits.size(); ++i)
-//    {
-//        const vector<const MateHit*>& m_hits = hits[i].mate_hits();
-//        for (size_t j = 0; j < m_hits.size(); ++j)
-//        {
-//            if (m_hits[j]->left_alignment() && !m_hits[j]->left_alignment()->contiguous())
-//            {
-//                spliced_hits.push_back(Scaffold(MateHit(m_hits[j]->ref_id(),
-//                                                        m_hits[j]->left_alignment(),
-//                                                        shared_ptr<const ReadHit>(),
-//                                                        0,
-//                                                        0)));
-//
-//            }
-//            if (m_hits[j]->right_alignment() && !m_hits[j]->right_alignment()->contiguous())
-//            {
-//                spliced_hits.push_back(Scaffold(MateHit(m_hits[j]->ref_id(),
-//                                                        m_hits[j]->right_alignment(),
-//                                                        shared_ptr<const ReadHit>(),
-//                                                        0,
-//                                                        0)));
-//                
-//            }
-//        }
-//    }
-//    
-//	Scaffold::merge(spliced_hits, smashed_gene, false);
-//	vector<bool> constitutive_introns(intron_doc.size(), true);
-//	
-//    vector<pair<int, int> > gaps = smashed_gene.gaps();
-//    
-//	//size_t intron_idx = 0;
-//    
-//	for(map<pair<int, int>, int >::const_iterator itr = intron_doc.begin();
-//		itr != intron_doc.end(); 
-//		++itr)
-//	{
-//		int i_left = itr->first.first;
-//		int i_right = itr->first.second;
-//		
-//        double cumul_cov = 0;
-//        for (size_t i = 0; i < i_right - i_left; ++i)
-//        {
-//            size_t pos = (i_left - bundle_left) + i;
-//            cumul_cov += depth_of_coverage[pos];
-//        }
-//        cumul_cov /= i_right - i_left;
-//#if ASM_VERBOSE
-//        fprintf(stderr, "retained intron %d-%d depth of cov: %lf\n", i_left, i_right, cumul_cov);
-//#endif
-//        if (cumul_cov / bundle_avg_doc >= pre_mrna_fraction)
-//        {
-//            continue;
-//        }
-//        
-//        for (size_t j = 0; j < hits.size(); ++j)
-//        {
-//            //if (hits[j].has_intron())
-//            //    continue;
-//            if (hits[j].match_length(i_left, i_right))
-//            {
-//                toss[j] = true;
-//            }
-//		}
-//	}
-//    
-//	for (size_t j = 0; j < hits.size(); ++j)
-//	{	
-//		if (!toss[j])
-//		{
-//			filtered_hits.push_back(hits[j]);
-//			//#if ASM_VERBOSE
-//			//			if (hits[j].has_intron())
-//			//			{
-//			//				
-//			//				fprintf(stderr, "KEEPING intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
-//			//			}
-//			//#endif	
-//		}
-//		else
-//		{
-//#if ASM_VERBOSE
-//			//			if (hits[j].has_intron())
-//			//			{
-//			//				
-//			//				fprintf(stderr, "\tFiltering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
-//			//			}
-//#endif	
-//		}
-//	}
-//	
-//	//#if ASM_VERBOSE
-//	//	fprintf(stderr, "\tInitial filter pass complete\n");
-//	//#endif
-//	
-//	hits = filtered_hits;
-//}
-
 void filter_hits(int bundle_length,
 				 int bundle_left,
 				 vector<Scaffold>& hits)
@@ -456,7 +320,7 @@ void filter_hits(int bundle_length,
 	
 	pre_mrna_filter(bundle_length, bundle_left, hits);
 	
-	vector<int> depth_of_coverage(bundle_length,0);
+	vector<int> depth_of_coverage(bundle_length+1,0);
 	vector<double> scaff_doc;
 	map<pair<int,int>, int> intron_doc;
 	vector<Scaffold> filtered_hits;
@@ -501,6 +365,10 @@ void filter_hits(int bundle_length,
 		{
 			for (size_t j = 0; j < hits.size(); ++j)
 			{
+				if (hits[j].is_ref())
+                {
+					continue;
+                }
 				int i_left = itr->first.first;
 				int i_right = itr->first.second;
 				int j_match_len = hits[j].match_length(i_left, i_right); 
@@ -587,7 +455,7 @@ void filter_hits(int bundle_length,
 		if (!toss[j])
 		{
 			filtered_hits.push_back(hits[j]);
-//#if ASM_VERBOSE
+//#if verbose_msg
 //			if (hits[j].has_intron())
 //			{
 //				
@@ -600,12 +468,12 @@ void filter_hits(int bundle_length,
 			if (hits[j].has_intron())
 			{
 				
-				asm_verbose("\t!!!Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
+				verbose_msg("\t!!!Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
 			}
 		}
 	}
 	
-//#if ASM_VERBOSE
+//#if verbose_msg
 //	fprintf(stderr, "\tInitial filter pass complete\n");
 //#endif
 	
@@ -623,7 +491,7 @@ void filter_hits(int bundle_length,
 								 dummy,
 								 false);
 	
-//#if ASM_VERBOSE
+//#if verbose_msg
 //	fprintf(stderr, "\tUpdated avg bundle doc = %lf\n", bundle_avg_doc);
 //#endif
 	
@@ -635,7 +503,7 @@ void filter_hits(int bundle_length,
 	
 	
 	
-//#if ASM_VERBOSE
+//#if verbose_msg
 //    double bundle_thresh = pre_mrna_fraction * bundle_avg_doc;
 //	fprintf(stderr, "\tthreshold is = %lf\n", bundle_thresh);
 //#endif
@@ -664,7 +532,7 @@ void filter_hits(int bundle_length,
 		if (!toss[j])
 		{
 			filtered_hits.push_back(hits[j]);
-//#if ASM_VERBOSE
+//#if verbose_msg
 //			if (hits[j].has_intron())
 //			{
 //				
@@ -676,7 +544,7 @@ void filter_hits(int bundle_length,
 		{
 			if (hits[j].has_intron())
 			{
-				asm_verbose("\t***Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
+				verbose_msg("\t***Filtering intron scaff [%d-%d]\n", hits[j].left(), hits[j].right());
 			}
 		}
 	}
@@ -688,7 +556,9 @@ void filter_hits(int bundle_length,
 
 
 void filter_junk_isoforms(vector<shared_ptr<Abundance> >& transcripts,
-						  vector<double>& abundances)
+						  vector<double>& abundances,
+                          const vector<shared_ptr<Abundance> >& mapped_transcripts,
+                          double locus_mass)
 {
 	//	vector<double>::iterator max_ab = std::max_element(abundances.begin(),
 	//													   abundances.end());
@@ -716,14 +586,12 @@ void filter_junk_isoforms(vector<shared_ptr<Abundance> >& transcripts,
 	vector<bool> repeats(transcripts.size(), false); // too many low-quality hits
 	vector<bool> too_rare(transcripts.size(), false); // too rare to be reliably quantitated, could be error
 	
-	vector<bool> illegal_microexon(transcripts.size(), false); // initial or terminal exons are too short
-	
 	//cerr << "Chucked : ";
 	for (size_t t = 0; t < transcripts.size(); ++t)
 	{
 		shared_ptr<Scaffold> scaff = transcripts[t]->transfrag();
-
-		if (allow_junk_filtering)
+        
+		if (!(scaff->is_ref()) && allow_junk_filtering)
 		{
 			const vector<const MateHit*> hits = scaff->mate_hits();
 			
@@ -745,32 +613,45 @@ void filter_junk_isoforms(vector<shared_ptr<Abundance> >& transcripts,
 				}
 			}
 			
-			double low_qual_hits = 0.0;
-			static const double low_qual_err_prob = high_phred_err_prob; // hits with error_prob() above this are low quality;
-			static const double low_qual_thresh = 0.75; // hits with more than this fraction of low qual hits are repeats
-			for (vector<const MateHit*>::const_iterator itr = hits.begin();
-				 itr != hits.end();
-				 ++itr)
-			{
-				double e = (*itr)->error_prob();
-				if (e >= low_qual_err_prob)
-					low_qual_hits += 1.0;
-			}
-			double low_qual_frac = low_qual_hits / (double)hits.size();
-			if (low_qual_frac > low_qual_thresh)
-				repeats[t] = true;
+            if (library_type != "transfrags")
+            {
+                double low_qual_hits = 0.0;
+                static const double low_qual_err_prob = high_phred_err_prob; // hits with error_prob() above this are low quality;
+                static const double low_qual_thresh = 0.75; // hits with more than this fraction of low qual hits are repeats
+                for (vector<const MateHit*>::const_iterator itr = hits.begin();
+                     itr != hits.end();
+                     ++itr)
+                {
+                    double e = 1-(*itr)->mass();
+                    if (e >= low_qual_err_prob)
+                        low_qual_hits += 1.0;
+                }
+            
+                double low_qual_frac = low_qual_hits / (double)hits.size();
+                if (low_qual_frac > low_qual_thresh)
+                    repeats[t] = true;
+            }
+            if (scaff->strand() == CUFF_FWD &&
+                (abundances[t] / max_fwd_ab) < min_isoform_fraction)
+                too_rare[t] = true;
+            if ((scaff->strand() == CUFF_REV ||  scaff->strand() == CUFF_STRAND_UNKNOWN) &&
+                (abundances[t] / max_rev_ab) < min_isoform_fraction)
+                too_rare[t] = true;
+
+            const vector<double>* cond_probs = (mapped_transcripts[t]->cond_probs());
+            if (cond_probs)
+            {
+                assert (library_type != "transfrags");
+                double supporting_hits = abundances[t] * locus_mass;
+                if (supporting_hits < min_frags_per_transfrag)
+                    chaff[t] = true;
+            }
 		}
-		
-		if (scaff->strand() == CUFF_FWD &&
-			(abundances[t] / max_fwd_ab) < min_isoform_fraction)
-			too_rare[t] = true;
-		if ((scaff->strand() == CUFF_REV ||  scaff->strand() == CUFF_STRAND_UNKNOWN) &&
-			(abundances[t] / max_rev_ab) < min_isoform_fraction)
-			too_rare[t] = true;
-        
-        if ((scaff->mate_hits().size() < min_frags_per_transfrag))
-            chaff[t] = true;
-		
+        else // we should still filter things that are zero to improve robustness of MAP estimation
+        {
+            if (abundances[t] == 0.0)
+                too_rare[t] = true;
+        }
 	}
 	
 	vector<shared_ptr<Abundance> > non_junk_transcripts;
@@ -784,7 +665,7 @@ void filter_junk_isoforms(vector<shared_ptr<Abundance> >& transcripts,
 		}
         else
         {
-            asm_verbose( "Filtering isoform %d-%d\n", transcripts[t]->transfrag()->left(), transcripts[t]->transfrag()->right());
+            verbose_msg( "Filtering isoform %d-%d\n", transcripts[t]->transfrag()->left(), transcripts[t]->transfrag()->right());
         }
 	}
 	
@@ -808,10 +689,19 @@ void filter_junk_genes(vector<Gene>& genes)
 	for (size_t i = 0; i < genes.size(); ++i)
 	{
 		const Gene& g = genes[i];
+		
+		if(g.has_ref_trans())
+		{
+			good_genes.push_back(g);
+			continue;
+		}
+		
 		bool good_gene = true;
 		for (size_t j = 0; j < all_isoforms.size(); ++j)
 		{
 			vector<pair<int, int> > introns = all_isoforms[j].scaffold().gaps();
+            
+            //assert (!allow_junk_filtering || all_isoforms[j].scaffold().mate_hits().size() >= min_frags_per_transfrag);
 			for (size_t k = 0; k < introns.size(); ++k)
 			{
 				if (g.left() > introns[k].first && g.right() < introns[k].second &&
@@ -834,10 +724,424 @@ void filter_junk_genes(vector<Gene>& genes)
         }
         else
         {
-            asm_verbose("Filtering transfrags from gene %d-%d\n", g.left(), g.right());
+            verbose_msg("Filtering transfrags from gene %d-%d\n", g.left(), g.right());
         }
 	}
 	
 	genes = good_genes;
 	
 }
+
+void clip_by_3_prime_dropoff(vector<Scaffold>& scaffolds)
+{
+    vector<pair<double, Scaffold*> > three_prime_ends;
+    
+    if (library_type != "transfrags")
+    {
+        foreach (Scaffold& scaff, scaffolds)
+        {
+            if (!(scaff.strand() == CUFF_FWD || scaff.strand() == CUFF_REV))
+                continue;
+
+            int scaff_len = scaff.length();
+            vector<double> coverage(scaff_len, 0.0);
+            
+            double total = 0;
+            foreach(const MateHit* hit, scaff.mate_hits())
+            {
+                int start, end, frag_len;
+                if (!scaff.map_frag(*hit, start, end, frag_len)) continue;
+                
+                if (scaff.strand() == CUFF_REV)
+                {
+                    start = scaff_len - 1 - start;
+                    end = scaff_len - 1 - end;
+                    swap(start, end);
+                }
+                
+                for(int i = start; i <= end; ++i)
+                {
+                    coverage[i] += hit->mass();
+                    total += hit->mass();
+                }
+            }
+            double avg_cov = total/scaff_len;
+    //        if (avg_cov < trim_3_avgcov_thresh)
+    //            continue;
+            
+            const AugmentedCuffOp* exon_3 = NULL;
+            int mult;
+            int offset;
+            
+            if (scaff.strand() == CUFF_REV)
+            {
+                mult = 1;
+                offset = 0;
+                exon_3 = &scaff.augmented_ops().front();
+            }
+            else if (scaff.strand() == CUFF_FWD)
+            {
+                mult = -1;
+                offset = scaff_len - 1;
+                exon_3 = &scaff.augmented_ops().back();
+            }
+            else
+            {
+                continue;
+            }
+
+            int to_remove;
+            double min_cost = numeric_limits<double>::max();
+            double mean_to_keep = 0.0;
+            double mean_to_trim = 0.0;
+            double tmp_mean_to_trim = 0.0;
+            double tmp_mean_to_keep = 0.0;
+            double tmp_mean_3prime = 0.0;
+            for (int i = 0; i < exon_3->genomic_length; i++)
+            {
+                tmp_mean_3prime += coverage[offset + mult*i];
+            }
+            tmp_mean_3prime /= exon_3->genomic_length;
+            
+            double base_cost = 0.0;
+            for (int i = 0; i < exon_3->genomic_length; i++)
+            {
+                double d = (coverage[offset + mult*i] - tmp_mean_3prime);
+                d *= d;
+                base_cost += d;
+            }
+            base_cost /= exon_3->genomic_length;
+            
+            size_t min_cost_x = -1;
+            for (to_remove = 1; to_remove < exon_3->genomic_length - 1; to_remove++)
+            {
+                tmp_mean_to_trim = 0.0;
+                tmp_mean_to_keep = 0.0;
+                for (size_t i = 0; i < exon_3->genomic_length; i++)
+                {
+                    if (i <= to_remove)
+                    {
+                        tmp_mean_to_trim += coverage[offset + mult*i];
+                    }
+                    else 
+                    {
+                        tmp_mean_to_keep += coverage[offset + mult*i];
+                    }
+                }
+                
+                tmp_mean_to_trim /= to_remove;
+                tmp_mean_to_keep /= (exon_3->genomic_length - to_remove);
+                
+                double tmp_mean_trim_cost = 0.0;
+                double tmp_mean_keep_cost = 0.0;
+                for (int i = 0; i < exon_3->genomic_length; i++)
+                {
+                    if (i <= to_remove)
+                    {
+                        double d = (coverage[offset + mult*i] - tmp_mean_to_trim);
+                        d *= d;
+                        tmp_mean_trim_cost += d;
+                    }
+                    else 
+                    {
+                        double d = (coverage[offset + mult*i] - tmp_mean_to_keep);
+                        d *= d;
+                        tmp_mean_keep_cost += d;
+                    }
+                }
+                
+                tmp_mean_trim_cost /= to_remove;
+                tmp_mean_keep_cost /= (exon_3->genomic_length - to_remove);
+                
+                double new_cost = tmp_mean_trim_cost + tmp_mean_keep_cost;
+                
+                if (new_cost < min_cost && trim_3_dropoff_frac * tmp_mean_to_keep > tmp_mean_to_trim && new_cost < base_cost && to_remove > scaff_len * 0.05)
+                {
+                    min_cost = tmp_mean_trim_cost + tmp_mean_keep_cost;
+                    min_cost_x = to_remove;
+                    mean_to_keep = tmp_mean_to_keep;
+                    mean_to_trim = tmp_mean_to_trim;
+                }
+            }
+            
+            // If trimming reduces the overall mean squared error of the coverage
+            // do it
+            if (avg_cov >= trim_3_avgcov_thresh && min_cost_x < exon_3->genomic_length)
+            {
+                scaff.trim_3(min_cost_x);
+            }
+            
+            // store the mean squared error for this exon
+            tmp_mean_3prime = 0.0;
+            for (int i = 0; i < exon_3->genomic_length; i++)
+            {
+                tmp_mean_3prime += coverage[offset + mult*i];
+            }
+            tmp_mean_3prime /= exon_3->genomic_length;
+            
+            base_cost = 0.0;
+            for (int i = 0; i < exon_3->genomic_length; i++)
+            {
+                double d = (coverage[offset + mult*i] - tmp_mean_3prime);
+                d *= d;
+                base_cost += d;
+            }
+            base_cost /= exon_3->genomic_length;
+            three_prime_ends.push_back(make_pair(base_cost, &scaff));
+        }
+    }
+    else
+    {
+        foreach (Scaffold& scaff, scaffolds)
+        {
+            if (!(scaff.strand() == CUFF_FWD || scaff.strand() == CUFF_REV))
+                continue;
+            
+            int scaff_len = scaff.length();
+            vector<double> coverage(scaff_len, 0.0);
+            
+            double total = 0;
+            foreach(const MateHit* hit, scaff.mate_hits())
+            {
+                int start, end, frag_len;
+                if (!scaff.map_frag(*hit, start, end, frag_len)) continue;
+                
+                if (scaff.strand() == CUFF_REV)
+                {
+                    start = scaff_len - 1 - start;
+                    end = scaff_len - 1 - end;
+                    swap(start, end);
+                }
+                
+                for(int i = start; i <= end; ++i)
+                {
+                    coverage[i] += hit->mass();
+                    total += hit->mass();
+                }
+            }
+            double avg_cov = total/scaff_len;
+            //        if (avg_cov < trim_3_avgcov_thresh)
+            //            continue;
+            
+            const AugmentedCuffOp* exon_3 = NULL;
+            int mult;
+            int offset;
+            
+            if (scaff.strand() == CUFF_REV)
+            {
+                mult = 1;
+                offset = 0;
+                exon_3 = &scaff.augmented_ops().front();
+            }
+            else if (scaff.strand() == CUFF_FWD)
+            {
+                mult = -1;
+                offset = scaff_len - 1;
+                exon_3 = &scaff.augmented_ops().back();
+            }
+            else
+            {
+                continue;
+            }
+            
+            three_prime_ends.push_back(make_pair(scaff.fpkm(), &scaff));
+        }
+        
+    }
+    
+    adjacency_list <vecS, vecS, undirectedS> G;
+	
+	for (size_t i = 0; i < three_prime_ends.size(); ++i)
+	{
+		add_vertex(G);
+	}
+	
+	for (size_t i = 0; i < three_prime_ends.size(); ++i)
+	{
+		Scaffold* scaff_i = three_prime_ends[i].second;
+		//assert (scaff_i);
+		
+        const AugmentedCuffOp* scaff_i_exon_3 = NULL;
+        
+        if (scaff_i->strand() == CUFF_REV)
+        {
+            scaff_i_exon_3 = &(scaff_i->augmented_ops().front());
+        }
+        else if (scaff_i->strand() == CUFF_FWD)
+        {
+            scaff_i_exon_3 = &(scaff_i->augmented_ops().back());
+        }
+        
+		for (size_t j = i + 1; j < three_prime_ends.size(); ++j)
+		{
+			Scaffold* scaff_j = three_prime_ends[j].second;
+            
+            if (scaff_i->strand() != scaff_j->strand())
+                continue;
+            
+            const AugmentedCuffOp* scaff_j_exon_3 = NULL;
+            
+            if (scaff_j->strand() == CUFF_REV)
+            {
+                scaff_j_exon_3 = &(scaff_j->augmented_ops().front());
+            }
+            else if (scaff_j->strand() == CUFF_FWD)
+            {
+                scaff_j_exon_3 = &(scaff_j->augmented_ops().back());
+            }
+			
+			if (AugmentedCuffOp::overlap_in_genome(*scaff_j_exon_3, *scaff_i_exon_3) && 
+                AugmentedCuffOp::compatible(*scaff_j_exon_3, *scaff_i_exon_3, 0))
+				add_edge(i, j, G);
+		}
+	}
+	
+	std::vector<int> component(num_vertices(G));
+	connected_components(G, &component[0]);
+	
+	vector<vector<bool> > clusters(three_prime_ends.size(), 
+								   vector<bool>(three_prime_ends.size(), false));
+	
+	//vector<vector<size_t> > cluster_indices(three_prime_ends.size());
+    
+    vector<vector<pair<double, Scaffold*> > > grouped_scaffolds(three_prime_ends.size());
+	for (size_t i = 0; i < three_prime_ends.size(); ++i)
+	{
+		clusters[component[i]][i] = true;
+		grouped_scaffolds[component[i]].push_back(three_prime_ends[i]);
+	}
+    
+    for (size_t i = 0; i < grouped_scaffolds.size(); ++i)
+    {
+        vector<pair<double, Scaffold*> >& group = grouped_scaffolds[i];
+        sort(group.begin(), group.end());
+        if (group.empty())
+            continue;
+        
+        Scaffold* group_leader = NULL;
+        int trim_point = -1;
+        
+        const AugmentedCuffOp* group_exon_3 = NULL;
+        vector<pair<double, Scaffold*> >::iterator l_itr = group.begin();
+        while (l_itr != group.end())
+        {
+            Scaffold* possible_leader = l_itr->second;
+            bool ok_clip_leader = true;
+            vector<pair<double, Scaffold*> >::iterator g_itr = group.begin();
+            const AugmentedCuffOp* l_exon_3 = NULL;
+            CuffStrand s = possible_leader->strand();
+
+            if (s != CUFF_STRAND_UNKNOWN)
+            {  
+                if (s == CUFF_REV)
+                    l_exon_3 = &(possible_leader->augmented_ops().front());
+                else 
+                    l_exon_3 = &(possible_leader->augmented_ops().back());
+                for (; g_itr != group.end(); ++g_itr)
+                {
+                    const AugmentedCuffOp* g_exon_3 = NULL;
+                    if (s == CUFF_REV)
+                    {
+                        //  bad:
+                        //              leader  
+                        //    follower
+                        g_exon_3 = &(g_itr->second->augmented_ops().front());
+                        if (g_exon_3->g_right() <= l_exon_3->g_left())
+                            ok_clip_leader = false;
+                        
+                        // for meta-assembly libraries, don't ever allow clipping, just extension
+                        // bad:
+                        //             leader
+                        //         follower
+                        if (library_type == "transfrags" && 
+                            g_exon_3->g_left() < l_exon_3->g_left())
+                            ok_clip_leader = false;
+                    }
+                    else 
+                    {
+                        //  bad:
+                        //          follower  
+                        //  leader
+                        g_exon_3 = &(g_itr->second->augmented_ops().back());
+                        if (g_exon_3->g_left() >= l_exon_3->g_right())
+                            ok_clip_leader = false;
+                        
+                        // for meta-assembly libraries, don't ever allow clipping, just extension
+                        // bad:
+                        //             leader
+                        //                follower
+                        if (library_type == "transfrags" && 
+                            g_exon_3->g_right() > l_exon_3->g_right())
+                            ok_clip_leader = false;
+                    }
+                }
+            }
+            else
+            {
+                ok_clip_leader = false;
+            }
+            
+            if (ok_clip_leader)
+            {
+                if (s == CUFF_REV)
+                {
+                    if (trim_point == -1)
+                        trim_point = l_exon_3->g_left();
+                    else if (l_exon_3->g_left() < trim_point)
+                        ok_clip_leader = false;
+                }
+                else 
+                {
+                    if (trim_point == -1)
+                        trim_point = l_exon_3->g_right();
+                    else if (l_exon_3->g_right() > trim_point)
+                        ok_clip_leader = false;
+                }
+            }
+            
+            if (ok_clip_leader)
+            {
+                group_leader = possible_leader;
+                group_exon_3 = l_exon_3;
+                break;
+            }
+            ++l_itr;
+        }
+        
+        if (!group_leader || !group_exon_3)
+            continue;
+        
+        for (size_t j = 0; j < group.size(); ++j)
+        {
+            const AugmentedCuffOp* exon_3 = NULL;
+            int end_diff = 0;
+            if (group_leader->strand() == CUFF_REV)
+            {
+                exon_3 = &(group[j].second->augmented_ops().front());
+                end_diff = group_exon_3->g_left() - exon_3->g_left();
+            }
+            else 
+            {
+                exon_3 = &(group[j].second->augmented_ops().back());                
+                end_diff = exon_3->g_right() - group_exon_3->g_right();
+            }
+            
+            if (end_diff > 0)
+            {
+                // leader
+                //     follower
+                group[j].second->trim_3(end_diff);
+            }
+            else if (end_diff < 0)
+            {
+                //        leader
+                //   follower
+                group[j].second->extend_3(-end_diff);
+            }
+        }
+    }
+    
+    
+	return;
+	
+}
