@@ -30,6 +30,136 @@ boost::mutex RefSequenceTable::table_lock;
 
 int num_deleted = 0;
 
+void ReadHit::trim(int trimmed_length)
+{
+    bool antisense_aln = _sam_flag & 0x10;
+    
+    vector<CigarOp> new_cigar;
+    int new_left = 0;
+    
+    if (!antisense_aln)
+    {
+        int pos = _left;
+        new_left = _left;
+        int length = 0;
+		for (vector<CigarOp>::iterator i = _cigar.begin(); i < _cigar.end(); ++i)
+		{
+			const CigarOp& op = *i;
+            
+            if (length < trimmed_length)
+            {
+                switch(op.opcode)
+                {
+                    case REF_SKIP:
+                        //gaps_out.push_back(make_pair(pos, pos + op.length - 1));
+                        pos += op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case SOFT_CLIP:
+                        assert(false); // not sure if this case is right
+                        pos += op.length;
+                        length += op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case HARD_CLIP:
+                        new_cigar.push_back(op);
+                        break;
+                    case MATCH:
+                        if (length + op.length < trimmed_length)
+                        {
+                            pos += op.length;
+                            length += op.length;
+                            new_cigar.push_back(op);
+                        }
+                        else
+                        {
+                            new_cigar.push_back(CigarOp(MATCH, trimmed_length - length));
+                            pos += trimmed_length - length;
+                            length += trimmed_length - length;
+                        }
+                        break;
+                    case INS:
+                        assert(false); // not sure if this case is right
+                        pos -= op.length;
+                        length -= op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case DEL:
+                        assert(false); // not sure if this case is right
+                        pos += op.length;
+                        length += op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    default:
+                        break;
+                }
+            }
+		}
+    }
+    else
+    {
+        int pos = _right;
+        int length = 0;
+		for (vector<CigarOp>::reverse_iterator i = _cigar.rbegin(); i < _cigar.rend(); ++i)
+		{
+			const CigarOp& op = *i;
+            
+            if (length < trimmed_length)
+            {
+                switch(op.opcode)
+                {
+                    case REF_SKIP:
+                        //gaps_out.push_back(make_pair(pos, pos + op.length - 1));
+                        pos -= op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case SOFT_CLIP:
+                        assert(false); // not sure if this case is right
+                        pos -= op.length;
+                        length += op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case HARD_CLIP:
+                        new_cigar.push_back(op);
+                        break;
+                    case MATCH:
+                        if (length + op.length < trimmed_length)
+                        {
+                            pos -= op.length;
+                            length += op.length;
+                            new_cigar.push_back(op);
+                        }
+                        else
+                        {
+                            new_cigar.push_back(CigarOp(MATCH, trimmed_length - length));
+                            pos -= trimmed_length - length;
+                            length += trimmed_length - length;
+                        }
+                        break;
+                    case INS:
+                        assert(false); // not sure if this case is right
+                        pos += op.length;
+                        length -= op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    case DEL:
+                        assert(false); // not sure if this case is right
+                        pos -= op.length;
+                        length += op.length;
+                        new_cigar.push_back(op);
+                        break;
+                    default:
+                        break;
+                }
+            }
+		}
+        _left = pos;
+    }
+    _cigar = new_cigar;
+    _right = get_right();
+    assert (trimmed_length == read_len());
+}
+
 //static const int max_read_length = 1024;
 
 bool hit_insert_id_lt(const ReadHit& h1, const ReadHit& h2)
@@ -46,6 +176,24 @@ bool hits_eq_mod_id(const ReadHit& lhs, const ReadHit& rhs)
 			lhs.cigar() == rhs.cigar());
 }
 
+// Compares for structural equality, but won't declare multihits equal to one another
+bool hits_eq_non_multi(const MateHit& lhs, const MateHit& rhs)
+{
+	if ((lhs.is_multi() || rhs.is_multi() ) && lhs.insert_id() != rhs.insert_id())
+		return false;
+	return hits_equals(lhs, rhs);
+}
+
+// Compares for structural equality, but won't declare multihits equal to one another
+// and won't return true for hits from different read groups (e.g. replicate samples)
+bool hits_eq_non_multi_non_replicate(const MateHit& lhs, const MateHit& rhs)
+{
+	if ((lhs.is_multi() || rhs.is_multi() || lhs.read_group_props() != rhs.read_group_props()) && lhs.insert_id() != rhs.insert_id())
+		return false;
+	return hits_equals(lhs, rhs);
+}
+    
+// Does NOT care about the read group this hit came from.
 bool hits_equals(const MateHit& lhs, const MateHit& rhs) 
 {
 	if (lhs.ref_id() != rhs.ref_id())
@@ -74,13 +222,14 @@ bool has_no_collapse_mass(const MateHit& hit)
 }
 
 // Assumes hits are sorted by mate_hit_lt
+// Does not collapse hits that are multi-reads
 void collapse_hits(const vector<MateHit>& hits,
 				   vector<MateHit>& non_redundant)
 {
 	copy(hits.begin(), hits.end(), back_inserter(non_redundant));
 	vector<MateHit>::iterator new_end = unique(non_redundant.begin(), 
 											   non_redundant.end(), 
-											   hits_equals);
+											   hits_eq_non_multi_non_replicate);
 	non_redundant.erase(new_end, non_redundant.end());
     non_redundant.resize(non_redundant.size());
 	
@@ -91,18 +240,26 @@ void collapse_hits(const vector<MateHit>& hits,
 	size_t curr_unique_aln = 0;
 	while (curr_aln < hits.size())
 	{
-		if (hits_equals(non_redundant[curr_unique_aln], hits[curr_aln]) || hits_equals(non_redundant[++curr_unique_aln], hits[curr_aln]))
-			non_redundant[curr_unique_aln].incr_collapse_mass(hits[curr_aln].mass());
+		if (hits_eq_non_multi_non_replicate(non_redundant[curr_unique_aln], hits[curr_aln]) || hits_eq_non_multi_non_replicate(non_redundant[++curr_unique_aln], hits[curr_aln]))
+		{
+            double more_mass = hits[curr_aln].common_scale_mass();
+			//assert(non_redundant[curr_unique_aln].collapse_mass() == 0 || !non_redundant[curr_unique_aln].is_multi());
+			non_redundant[curr_unique_aln].incr_collapse_mass(more_mass);
+		}
 		else
 			assert(false);
 		
 		++curr_aln;
 	}
 	
-	non_redundant.erase(remove_if(non_redundant.begin(),non_redundant.end(),has_no_collapse_mass), non_redundant.end()); 
+	//foreach(MateHit& hit, non_redundant)
+		//assert(hit.collapse_mass() <= 1 || !hit.is_multi());
+	
+	//non_redundant.erase(remove_if(non_redundant.begin(),non_redundant.end(),has_no_collapse_mass), non_redundant.end()); 
 	
 }
 
+// Places multi-reads to the right of reads they match
 bool mate_hit_lt(const MateHit& lhs, const MateHit& rhs)
 {
 	if (lhs.left() != rhs.left())
@@ -151,6 +308,11 @@ bool mate_hit_lt(const MateHit& lhs, const MateHit& rhs)
 		}
 	}
 	
+	if (lhs.is_multi() != rhs.is_multi())
+	{
+		return rhs.is_multi();
+	}
+	
 	return false;
 }
 
@@ -159,54 +321,58 @@ ReadHit HitFactory::create_hit(const string& insert_name,
 							   const string& ref_name,
 							   int left,
 							   const vector<CigarOp>& cigar,
-							   bool antisense_aln,
 							   CuffStrand source_strand,
 							   const string& partner_ref,
 							   int partner_pos, 
-							   double error_prob,
-							   unsigned int edit_dist)
+							   unsigned int edit_dist,
+							   int num_hits,
+                               float base_mass,
+                               uint32_t sam_flag)
 {
-	uint64_t insert_id = _insert_table.get_id(insert_name);
-	uint32_t reference_id = _ref_table.get_id(ref_name, NULL);
-	uint32_t partner_ref_id = _ref_table.get_id(partner_ref, NULL);
+	InsertID insert_id = _insert_table.get_id(insert_name);
+	RefID reference_id = _ref_table.get_id(ref_name, NULL);
+	RefID partner_ref_id = _ref_table.get_id(partner_ref, NULL);
 	
 	return ReadHit(reference_id,
-				   insert_id, 
+				   insert_id,
 				   left, 
 				   cigar, 
-				   antisense_aln,
 				   source_strand,
 				   partner_ref_id,
 				   partner_pos,
-				   error_prob,
-				   edit_dist);	
+				   edit_dist,
+				   num_hits,
+                   base_mass,
+                   sam_flag);	
 }
 
 ReadHit HitFactory::create_hit(const string& insert_name, 
 							   const string& ref_name,
 							   uint32_t left,
 							   uint32_t read_len,
-							   bool antisense_aln,
 							   CuffStrand source_strand,
 							   const string& partner_ref,
 							   int partner_pos,
-							   double error_prob,
-							   unsigned int edit_dist)
+							   unsigned int edit_dist,
+							   int num_hits,
+                               float base_mass,
+                               uint32_t sam_flag)
 {
-	uint64_t insert_id = _insert_table.get_id(insert_name);
-	uint32_t reference_id = _ref_table.get_id(ref_name, NULL);
-	uint32_t partner_ref_id = _ref_table.get_id(partner_ref, NULL);
+	InsertID insert_id = _insert_table.get_id(insert_name);
+	RefID reference_id = _ref_table.get_id(ref_name, NULL);
+	RefID partner_ref_id = _ref_table.get_id(partner_ref, NULL);
 	
 	return ReadHit(reference_id,
 				   insert_id, 
 				   left,
 				   read_len,
-				   antisense_aln,
 				   source_strand,
 				   partner_ref_id,
 				   partner_pos,
-				   error_prob,
-				   edit_dist);	
+				   edit_dist,
+				   num_hits,
+                   base_mass,
+                   sam_flag);	
 }
 
 // populate a bam_t This will 
@@ -237,9 +403,10 @@ bool BAMHitFactory::next_record(const char*& buf, size_t& buf_size)
 	return true;
 }
 
-CuffStrand use_stranded_protocol(uint32_t sam_flag, bool antisense_aln, MateStrandMapping msm)
+CuffStrand use_stranded_protocol(uint32_t sam_flag,  MateStrandMapping msm)
 {
-	if (((sam_flag & BAM_FPAIRED) && (sam_flag & BAM_FREAD1)) || !(sam_flag & BAM_FPAIRED)) // left-most read or single-end
+    bool antisense_aln = sam_flag & 0x10;
+	if (((sam_flag & BAM_FPAIRED) && (sam_flag & BAM_FREAD1)) || !(sam_flag & BAM_FPAIRED)) // first-in-pair or single-end
 	{
 		switch(msm)
 		{
@@ -253,7 +420,7 @@ CuffStrand use_stranded_protocol(uint32_t sam_flag, bool antisense_aln, MateStra
 				break;
 		}
 	}
-	else // right-most read
+	else // second-in-pair read
 	{
 		switch (msm)
 		{
@@ -289,34 +456,24 @@ bool BAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 	
 	vector<CigarOp> cigar;
 	bool spliced_alignment = false;
+	int num_hits = 1;
 	
-	double mapQ = hit_buf->core.qual;
-	long double error_prob;
-	if (mapQ > 0)
-	{
-		long double p = (-1.0 * mapQ) / 10.0;
-		error_prob = pow(10.0L, p);
-	}
-	else
-	{
-		error_prob = 1.0;
-	}
-
 	//header->target_name[c->tid]
 	
-	if (target_id < 0)
+	if (sam_flag & 0x4 || target_id < 0)
 	{
 		//assert(cigar.size() == 1 && cigar[0].opcode == MATCH);
 		bh = create_hit(bam1_qname(hit_buf),
 						"*",
 						0, // SAM files are 1-indexed
 						0,
-						false,
 						CUFF_STRAND_UNKNOWN,
 						"*",
 						0,
-						1.0,
-						0);
+						0,
+						1,
+                        1.0,
+                        sam_flag);
 		return true;
 	}
 	
@@ -329,7 +486,7 @@ bool BAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 		int length = bam1_cigar(hit_buf)[i] >> BAM_CIGAR_SHIFT;
 		if (length <= 0)
 		{
-			fprintf (stderr, "BAM error: CIGAR op has zero length\n");
+		  fprintf (stderr, "BAM error: CIGAR op has zero length (%s)\n", bam1_qname(hit_buf));
 			return false;
 		}
 		
@@ -365,11 +522,11 @@ bool BAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 		if (mate_target_id == target_id)
 		{
 			mrnm = _hit_file->header->target_name[mate_target_id];
-			if (abs((int)text_mate_pos - (int)text_offset) > (int)max_intron_length)
-			{
-				//fprintf (stderr, "Mates are too distant, skipping\n");
-				return false;
-			}
+//			if (abs((int)text_mate_pos - (int)text_offset) > (int)max_intron_length)
+//			{
+//				//fprintf (stderr, "Mates are too distant, skipping\n");
+//				return false;
+//			}
 		}
 		else
 		{
@@ -401,25 +558,40 @@ bool BAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 		num_mismatches = bam_aux2i(ptr);
 	}
 
-    bool antisense_aln = bam1_strand(hit_buf);
+	ptr = bam_aux_get(hit_buf, "NH");
+	if (ptr)
+	{
+		num_hits = bam_aux2i(ptr);
+	}
     
+    double mass = 1.0;
+    ptr = bam_aux_get(hit_buf, "ZF");
+	if (ptr)
+	{
+		mass = bam_aux2i(ptr);
+        if (mass <= 0.0)
+            mass = 1.0;
+	}
+	    
     if (_rg_props.strandedness() == STRANDED_PROTOCOL && source_strand == CUFF_STRAND_UNKNOWN)
-		source_strand = use_stranded_protocol(sam_flag, antisense_aln, _rg_props.mate_strand_mapping());
+		source_strand = use_stranded_protocol(sam_flag, _rg_props.mate_strand_mapping());
     
 	if (!spliced_alignment)
 	{
-		
+		//assert(_rg_props.strandedness() == STRANDED_PROTOCOL || source_strand == CUFF_STRAND_UNKNOWN);
+
 		//assert(cigar.size() == 1 && cigar[0].opcode == MATCH);
 		bh = create_hit(bam1_qname(hit_buf),
 						text_name,
 						text_offset,  // BAM files are 0-indexed
 						cigar,
-						antisense_aln,
 						source_strand,
 						mrnm,
 						text_mate_pos,
-						error_prob,
-						num_mismatches);
+						num_mismatches,
+						num_hits,
+                        mass,
+                        sam_flag);
 		return true;
 		
 	}
@@ -434,12 +606,13 @@ bool BAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 						text_name,
 						text_offset,  // BAM files are 0-indexed
 						cigar,
-						antisense_aln,
 						source_strand,
 						mrnm,
 						text_mate_pos,
-						error_prob,
-						num_mismatches);
+						num_mismatches,
+						num_hits,
+                        mass,
+                        sam_flag);
 		return true;
 	}
 	
@@ -512,12 +685,19 @@ bool HitFactory::parse_header_string(const string& header_rec,
                 // in an order consistent with the header, and to enforce that
                 // BAM records appear in the order implied by the header
                 RefID _id = _ref_table.get_id(fields[1], NULL);
-                
-                const RefSequenceTable::SequenceInfo* info = _ref_table.get_info(_id);
 
+                const RefSequenceTable::SequenceInfo* info = _ref_table.get_info(_id);
+				
                 if (info->observation_order != _num_seq_header_recs)
                 {
-                    fprintf(stderr, "Error: sort order of reads in BAMs must be the same\n");
+                    if (info->name != fields[1])
+                    {
+                        fprintf(stderr, "Error: Hash collision between references '%s' and '%s'.\n", info->name, fields[1].c_str());
+                    }
+                    else
+                    {
+                        fprintf(stderr, "Error: sort order of reads in BAMs must be the same\n");
+					}
                     exit(1);
                 }
             }
@@ -543,7 +723,7 @@ void HitFactory::finalize_rg_props()
     }
 }
 
-static const int MAX_HEADER_LEN = 4 * 1024 * 1024; // 4 MB
+static const unsigned MAX_HEADER_LEN = 4 * 1024 * 1024; // 4 MB
 
 bool BAMHitFactory::inspect_header()
 {
@@ -616,7 +796,8 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 									 char* name_out,
 									 char* name_tags)
 {	
-	char bwt_buf[2048];
+	char bwt_buf[10*2048];
+
 	strcpy(bwt_buf, orig_bwt_buf);
 	// Are we still in the header region?
 	if (bwt_buf[0] == '@')
@@ -691,32 +872,22 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 	//int len = strlen(sequence);
 	vector<CigarOp> cigar;
 	bool spliced_alignment = false;
-	
-	double mapQ = atoi(map_qual_str);
-	long double error_prob;
-	if (mapQ > 0)
-	{
-		long double p = (-1.0 * mapQ) / 10.0;
-		error_prob = pow(10.0L, p);
-	}
-	else
-	{
-		error_prob = 1.0;
-	}
-	
-	if (!strcmp(text_name, "*"))
+	int num_hits = 1;
+	    
+    if ((sam_flag & 0x4) ||!strcmp(text_name, "*"))
 	{
 		//assert(cigar.size() == 1 && cigar[0].opcode == MATCH);
 		bh = create_hit(name,
-						text_name,
+						"*",
 						0, // SAM files are 1-indexed
 						0,
-						false,
 						CUFF_STRAND_UNKNOWN,
 						"*",
 						0,
-						1.0,
-						0);
+						0,
+						1,
+                        1.0,
+                        sam_flag);
 		return true;
 	}
 	// Mostly pilfered direct from the SAM tools:
@@ -780,11 +951,11 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 		if (!strcmp(mate_ref_name, "=") || !strcmp(mate_ref_name, text_name))
 		{
 			mrnm = text_name;
-			if (abs((int)text_mate_pos - (int)text_offset) > (int)max_intron_length)
-			{
-				//fprintf (stderr, "Mates are too distant, skipping\n");
-				return false;
-			}
+//			if (abs((int)text_mate_pos - (int)text_offset) > (int)max_intron_length)
+//			{
+//				//fprintf (stderr, "Mates are too distant, skipping\n");
+//				return false;
+//			}
 		}
 		else
 		{
@@ -802,6 +973,8 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 	
 	const char* tag_buf = buf;
 	
+    double mass = 1.0;
+    
 	while((tag_buf = strsep((char**)&buf,"\t")))
 	{
 		
@@ -829,6 +1002,16 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 				{
 					num_mismatches = atoi(third_token);
 				}
+				else if (!strcmp(first_token, "NH"))
+				{
+                    num_hits = atoi(third_token);
+				}
+                else if (!strcmp(first_token, "ZF"))
+				{
+					mass = atof(third_token);
+                    if (mass <= 0.0)
+                        mass = 1.0;
+				}
 				else 
 				{
 					
@@ -836,27 +1019,25 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 			}
 		}
 	}
-    
-    bool antisense_aln = sam_flag & 0x0010;
-    
+        
     // Don't let the protocol setting override explicit XS tags
 	if (_rg_props.strandedness() == STRANDED_PROTOCOL && source_strand == CUFF_STRAND_UNKNOWN)
-		source_strand = use_stranded_protocol(sam_flag, antisense_aln, _rg_props.mate_strand_mapping());
+		source_strand = use_stranded_protocol(sam_flag, _rg_props.mate_strand_mapping());
 	
 	if (!spliced_alignment)
-	{
-		
+	{		
 		//assert(cigar.size() == 1 && cigar[0].opcode == MATCH);
 		bh = create_hit(name,
 						text_name,
 						text_offset - 1,
 						cigar,
-						antisense_aln,
 						source_strand,
 						mrnm,
 						text_mate_pos - 1,
-						error_prob,
-						num_mismatches);
+						num_mismatches,
+						num_hits,
+                        mass,
+                        sam_flag);
 		return true;
 		
 	}
@@ -871,12 +1052,13 @@ bool SAMHitFactory::get_hit_from_buf(const char* orig_bwt_buf,
 						text_name,
 						text_offset - 1,
 						cigar,
-						antisense_aln,
 						source_strand,
 						mrnm,
 						text_mate_pos - 1,
-						error_prob,
-						num_mismatches);
+						num_mismatches,
+						num_hits,
+                        mass,
+                        sam_flag);
 		return true;
 	}
 	return false;
